@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2024	PgPool Global Development Group
+ * Copyright (c) 2003-2023	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -43,6 +43,7 @@
 #include <stdlib.h>
 
 #include "pool.h"
+#include "context/pool_query_context.h"
 #include "utils/pool_stream.h"
 #include "utils/palloc.h"
 #include "pool_config.h"
@@ -67,8 +68,6 @@ static POOL_CONNECTION_POOL * new_connection(POOL_CONNECTION_POOL * p);
 static int	check_socket_status(int fd);
 static bool connect_with_timeout(int fd, struct addrinfo *walk, char *host, int port, bool retry);
 
-#define TMINTMAX 0x7fffffff
-
 /*
 * initialize connection pools. this should be called once at the startup.
 */
@@ -160,6 +159,13 @@ pool_get_cp(char *user, char *database, int protoMajor, int check_socket)
 					ereport(LOG,
 							(errmsg("connection closed."),
 							 errdetail("retry to create new connection pool")));
+					/*
+					 * It is possible that one of backend just broke.  sleep 1
+					 * second to wait for failover occurres, then wait for the
+					 * failover finishes.
+					 */
+					sleep(1);
+					wait_for_failover_to_finish();
 
 					for (j = 0; j < NUM_BACKENDS; j++)
 					{
@@ -249,7 +255,6 @@ pool_create_cp(void)
 	POOL_CONNECTION_POOL *oldestp;
 	POOL_CONNECTION_POOL *ret;
 	ConnectionInfo *info;
-	int		main_node_id;
 
 	POOL_CONNECTION_POOL *p = pool_connection_pool;
 
@@ -262,7 +267,7 @@ pool_create_cp(void)
 
 	for (i = 0; i < pool_config->max_pool; i++)
 	{
-		if (in_use_backend_id(p) < 0)	/* is this connection pool out of use? */
+		if (MAIN_CONNECTION(p) == NULL)
 		{
 			ret = new_connection(p);
 			if (ret)
@@ -280,25 +285,21 @@ pool_create_cp(void)
 	 * discard it.
 	 */
 	oldestp = p = pool_connection_pool;
-	closetime = TMINTMAX;
+	closetime = MAIN_CONNECTION(p)->closetime;
 	pool_index = 0;
 
 	for (i = 0; i < pool_config->max_pool; i++)
 	{
-		main_node_id = in_use_backend_id(p);
-		if (main_node_id < 0)
-			elog(ERROR, "no in use backend found");	/* this should not happen */
-
 		ereport(DEBUG1,
 				(errmsg("creating connection pool"),
 				 errdetail("user: %s database: %s closetime: %ld",
-						   CONNECTION_SLOT(p, main_node_id)->sp->user,
-						   CONNECTION_SLOT(p, main_node_id)->sp->database,
-						   CONNECTION_SLOT(p, main_node_id)->closetime)));
+						   MAIN_CONNECTION(p)->sp->user,
+						   MAIN_CONNECTION(p)->sp->database,
+						   MAIN_CONNECTION(p)->closetime)));
 
-		if (CONNECTION_SLOT(p, main_node_id)->closetime < closetime)
+		if (MAIN_CONNECTION(p)->closetime < closetime)
 		{
-			closetime = CONNECTION_SLOT(p, main_node_id)->closetime;
+			closetime = MAIN_CONNECTION(p)->closetime;
 			oldestp = p;
 			pool_index = i;
 		}
@@ -306,21 +307,18 @@ pool_create_cp(void)
 	}
 
 	p = oldestp;
-	main_node_id = in_use_backend_id(p);
-	if (main_node_id < 0)
-		elog(ERROR, "no in use backend found");	/* this should not happen */
 	pool_send_frontend_exits(p);
 
 	ereport(DEBUG1,
 			(errmsg("creating connection pool"),
 			 errdetail("discarding old %zd th connection. user: %s database: %s",
 					   oldestp - pool_connection_pool,
-					   CONNECTION_SLOT(p, main_node_id)->sp->user,
-					   CONNECTION_SLOT(p, main_node_id)->sp->database)));
+					   MAIN_CONNECTION(p)->sp->user,
+					   MAIN_CONNECTION(p)->sp->database)));
 
 	for (i = 0; i < NUM_BACKENDS; i++)
 	{
-		if (CONNECTION_SLOT(p, i) == NULL)
+		if (!VALID_BACKEND(i))
 			continue;
 
 		if (!freed)
@@ -401,6 +399,8 @@ pool_backend_timer_handler(int sig)
 void
 pool_backend_timer(void)
 {
+#define TMINTMAX 0x7fffffff
+
 	POOL_CONNECTION_POOL *p = pool_connection_pool;
 	int			i,
 				j;
@@ -876,7 +876,6 @@ static POOL_CONNECTION_POOL * new_connection(POOL_CONNECTION_POOL * p)
 	int			active_backend_count = 0;
 	int			i;
 	bool		status_changed = false;
-	volatile BACKEND_STATUS	status;
 
 	MemoryContext oldContext = MemoryContextSwitchTo(TopMemoryContext);
 
@@ -901,8 +900,8 @@ static POOL_CONNECTION_POOL * new_connection(POOL_CONNECTION_POOL * p)
 		 * that the local status is up, while the global status has been
 		 * changed to down by failover.
 		 */
-		status = BACKEND_INFO(i).backend_status;
-		if (status != CON_UP && status != CON_CONNECT_WAIT)
+		if (BACKEND_INFO(i).backend_status != CON_UP &&
+			BACKEND_INFO(i).backend_status != CON_CONNECT_WAIT)
 		{
 			ereport(DEBUG1,
 					(errmsg("creating new connection to backend"),
@@ -910,8 +909,7 @@ static POOL_CONNECTION_POOL * new_connection(POOL_CONNECTION_POOL * p)
 							   i, BACKEND_INFO(i).backend_status)));
 
 			/* sync local status with global status */
-			*(my_backend_status[i]) = status;
-			my_main_node_id = Req_info->main_node_id;
+			*(my_backend_status[i]) = BACKEND_INFO(i).backend_status;
 			continue;
 		}
 
@@ -1065,15 +1063,11 @@ close_all_backend_connections(void)
 
 	for (i = 0; i < pool_config->max_pool; i++, p++)
 	{
-		int	backend_id = in_use_backend_id(p);
-
-		if (backend_id < 0)
+		if (!MAIN_CONNECTION(p))
 			continue;
-		if (CONNECTION_SLOT(p, backend_id) == NULL)
+		if (!MAIN_CONNECTION(p)->sp)
 			continue;
-		if (CONNECTION_SLOT(p, backend_id)->sp == NULL)
-			continue;
-		if (CONNECTION_SLOT(p, backend_id)->sp->user == NULL)
+		if (MAIN_CONNECTION(p)->sp->user == NULL)
 			continue;
 		pool_send_frontend_exits(p);
 	}
@@ -1081,38 +1075,15 @@ close_all_backend_connections(void)
 	POOL_SETMASK(&oldmask);
 }
 
-/*
- * Return number of established connections in the connection pool.
- * This is called when a client disconnects to pgpool.
- */
-void
-update_pooled_connection_count(void)
+void update_pooled_connection_count(void)
 {
 	int i;
 	int count = 0;
 	POOL_CONNECTION_POOL *p = pool_connection_pool;
-	for (i = 0; i < pool_config->max_pool; i++, p++)
+	for (i = 0; i < pool_config->max_pool; i++)
 	{
 		if (MAIN_CONNECTION(p))
 			count++;
 	}
 	pool_get_my_process_info()->pooled_connections = count;
 }
-
-/*
- * Return the first node id in use.
- * If no node is in use, return -1.
- */
-int
-in_use_backend_id(POOL_CONNECTION_POOL *pool)
-{
-	int	i;
-
-	for (i = 0; i < NUM_BACKENDS; i++)
-	{
-		if (pool->slots[i])
-			return i;
-	}
-
-	return -1;
-}
